"""Server Management tools."""
# pylint: disable=superfluous-parens
# pylint: disable=no-member
# pylint: disable=protected-access
# pylint: disable=invalid-name
import json

from django.contrib import admin
from autobahn.twisted.websocket import (WebSocketServerProtocol,
                                        WebSocketServerFactory)

from elrados.backend.cache import Cache


class BroadcastServerProtocol(WebSocketServerProtocol):
    """"Protocol for a Websocket server which broadcasts messages to clients"""
    def onOpen(self):
        """"Register the client for future broadcasts."""
        super(BroadcastServerProtocol, self).onOpen()
        self.factory.register(self)

    def connectionLost(self, reason):
        """"Unregister the client from future broadcasts."""
        super(BroadcastServerProtocol, self).connectionLost(reason)
        self.factory.unregister(self)


class BroadcastServerFactory(WebSocketServerFactory):
    """Broadcast server factory of the websocket."""
    protocol = BroadcastServerProtocol

    def __init__(self, url, settings=None):
        super(BroadcastServerFactory, self).__init__(url=url)
        self.clients = set()
        self.cache = Cache(self)
        self.settings = settings if settings is not None else {}

    def initialize_resources(self, resources):
        """Initialize cache resources."""
        display_attrs = {
            resource.__name__: admin.site._registry[resource].list_display
            for resource in resources
            }
        self.cache.initialize_display_list_cache(display_attrs)
        self.cache.initialize_resources_cache(resources)

        # send to all users the extracted resources
        self.cache.update_users_display_list()
        self.cache.update_users_resources_cache()
        print("[Frontend] Initialize done!")

    def broadcast(self, message, isBinary):
        """Broadcast a message to the clients."""
        for client in self.clients:
            client.sendMessage(message, isBinary=isBinary)

    def register(self, client):
        """Register new user to the websocket and initialize its data."""
        self.clients.add(client)
        client.sendMessage(json.dumps(
            {
                "event_type": "initialize-settings",
                "content": self.settings
                }
            ), False)
        client.sendMessage(json.dumps(
            {
                "event_type": "initialize-display-list",
                "content": self.cache.display_list_cache
                }
            ), False)
        client.sendMessage(json.dumps(
            {
                "event_type": "initialize-cache",
                "content": self.cache.resources_cache
                }
            ), False)

    def unregister(self, client):
        """Unregister user from the websocket."""
        self.clients.remove(client)



"""Server cache data."""
# pylint: disable=superfluous-parens
# pylint: disable=no-member
# pylint: disable=protected-access
# pylint: disable=invalid-name
# pylint: disable=unused-argument

import json
import threading

from elrados.backend import utils


DELETE_ACTION_FLAG = 3


class Cache(object):
    """Cache of the server that stores all the data needed for the frontend."""
    def __init__(self, factory):
        self._factory = factory
        self.resources_cache = {}
        self.display_list_cache = {}

    def initialize_display_list_cache(self, display_list):
        """Initialize display list cache."""
        print("[Frontend] initializing display list cache")
        self.display_list_cache = display_list

    def _insert_resource_to_cache(self, resource):
        """Add the given resource to the cache.

        Arguments:
            resource (django.db.models.Model): a model to save in the cache.
        """
        resource_name = str(resource.__name__)
        self.resources_cache[resource_name] = utils.get_resource_data(resource)

    def initialize_resources_cache(self, resources):
        """Initialize resources cache."""
        print("[Frontend] initializing resources cache")
        threads = []
        for resource in resources:
            resource_name = str(resource.__name__)
            if resource_name not in self.resources_cache:
                self.resources_cache[resource_name] = {}

            thread = threading.Thread(target=self._insert_resource_to_cache,
                                      args=(resource,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

    def update_users_resources_cache(self):
        """Send to all users the resources cache."""
        self._factory.broadcast(json.dumps(
            {
                "event_type": "initialize",
                "content": self.resources_cache
                }
            ), False)

    def update_users_display_list(self):
        """Send to all users the display list cache."""
        self._factory.broadcast(json.dumps(
            {
                "event_type": "initialize-display-list",
                "content": self.display_list_cache
                }
            ), False)

    def delete_from_resources_cache(self, object_id):
        """Delete data from the resources cache."""
        for resource_cache in self.resources_cache.values():
            if object_id in resource_cache:
                del resource_cache[object_id]
                break

    def update_resource(self, sender, resource, **kwargs):
        """Update a resource according to the resource update given."""
        del kwargs["signal"]
        resource_id = utils.get_object_id(resource)
        resource_name = str(utils.get_leaf(resource).__class__.__name__)
        print("[Frontend] Updating resource {}".format(resource_name))
        if resource_name == "LogEntry":
            if resource.action_flag == DELETE_ACTION_FLAG:
                self.delete_from_resources_cache(resource_id)

            self._factory.broadcast(json.dumps(
                {
                    "event_type": "resource_updated",
                    "content": {
                        resource_name: {
                            resource_id: utils.expand_resource(resource)
                        }
                    }
                }
                ), False)
            return

        if resource_name not in self.resources_cache:
            self.resources_cache[resource_name] = {}

        if resource_id not in self.resources_cache[resource_name]:
            self.resources_cache[resource_name][resource_id] = {}

        self.resources_cache[resource_name][resource_id].update(
            utils.expand_resource(resource))
        self._factory.broadcast(json.dumps(
            {
                "event_type": "resource_updated",
                "content": {
                    resource_name: {
                        resource_id:
                            self.resources_cache[resource_name][resource_id]
                    }
                }
            }
            ), False)

        print("[Frontend] Sent to all registered users")



"""Elrados web-socket server starter."""
# pylint: disable=superfluous-parens
# pylint: disable=no-member
# pylint: disable=protected-access
# pylint: disable=invalid-name
import os

import crochet
from twisted.internet import reactor
from django.db.models.signals import post_save
from django.contrib.auth.models import User, Group
from rotest.management import ResourceData
from rotest.common.config import SHELL_APPS
from rotest.management.utils.resources_discoverer import get_resources

from elrados.backend.management import BroadcastServerFactory

SERVER_PORT = 9000


class WebsocketService(object):
    """Websocket service that is exposed to the user of elrados."""
    def __init__(self, settings=None):
        print("[Frontend] initializing websocket service")
        self.factory = BroadcastServerFactory(
            u'ws://0.0.0.0:{}'.format(SERVER_PORT), settings)

    @crochet.run_in_reactor
    def create_server(self, init_list):
        """Create new user and listen to the server port."""
        self.factory.initialize_resources(init_list)
        reactor.listenTCP(SERVER_PORT, self.factory)

    def send_to_client(self, sender, instance, **kwargs):
        """Send update message to client."""
        self.factory.cache.update_resource(sender, instance, **kwargs)


def setup_server():
    """Start the websocket server."""
    crochet.setup()
    resource_models = [User, Group]
    for application in SHELL_APPS:
        resource_models.extend(resource.DATA_CLASS for resource in
                               get_resources(application).values()
                               if isinstance(resource.DATA_CLASS,
                                             ResourceData))

    if os.environ.get("RUN_MAIN") == "true":
        backend = WebsocketService(settings={
            "default_resource": resource_models[-1].__name__
        })
        backend.create_server(resource_models)
        post_save.connect(backend.send_to_client,
                          sender='management.ResourceData',
                          weak=False)

        for resource_data in resource_models:
            post_save.connect(backend.send_to_client,
                              sender=resource_data,
                              weak=False)



"""Backend utils module."""
import json

from django.db import models

from rotest.common.django_utils.common import get_fields, get_sub_model


def get_object_id(resource):
    """"Get the id of a given resource."""
    if hasattr(resource, "object_id"):
        return int(resource.object_id)

    return int(resource.id)


def get_leaf(model):
    """"Get the leaf resource data inheriting from model given.

    Arguments:
        model (django.db.models.Model): a model instance to get its leaf.

    Returns:
        django.db.models.Model. the leaf model of the given instance.
    """
    sub_model = get_sub_model(model)
    if sub_model is None:
        return model

    return get_leaf(sub_model)


def expand_resource(resource):
    """"Expand the given resource to it's fields.

    Arguments:
        resource (models.Model): a instance of a model to expand.

    Note:
        if a model has a reference to other model a `link` is created to
        the other model:

        Examples:
            {
               "id": model_id,
               "type": model_class_type
            }

    Returns:
        dict. the resource representation as a dict.
    """
    # refresh object from db.
    resource = resource.__class__.objects.get(id=resource.id)

    fields = get_fields(resource)
    for field_name, field_value in fields.items():
        if isinstance(field_value, models.Model):
            fields[field_name] = {
                "link": {
                    "id": get_object_id(field_value),
                    "type": get_leaf(field_value).__class__.__name__
                }
            }

        if isinstance(field_value, list):
            new_list = []
            for sub_element in field_value:
                if isinstance(sub_element, models.Model):
                    new_list.append({
                        "link": {
                            "id": get_object_id(sub_element),
                            "type": get_leaf(sub_element).__class__.__name__
                        }
                    })

                else:
                    new_list.append(sub_element)

        try:
            json.dumps(fields[field_name])

        except TypeError:
            fields[field_name] = str(fields[field_name])

    if hasattr(resource, "is_available"):
        fields["is_available"] = resource.is_available()

    return fields


def get_resource_data(resource_type):
    """Get all data of the given resource list from the db.

    Arguments:
        resource_type (django.db.models.Model): fetch all objects of a given
            resource.

    Returns:
        dict. all the objects of the given model.
            keys are the object id and values are dictionaries that represent
            the objects' datas.
    """
    resources = resource_type.objects.all()
    data_dict = {}
    for resource in resources:
        data_dict[get_object_id(resource)] = expand_resource(resource)

    return data_dict



"""Django definitions for the administrator site urls."""
from django.conf.urls import url

from elrados import utils

# pylint: disable=invalid-name
urlpatterns = [
    url(r'^api/elrados/release_owner/(?P<data_name>.*?)/?$',
        utils.set_field("owner"), name="remove_resource_owner"),
    url(r'^api/elrados/release_reserved/(?P<data_name>.*?)/?$',
        utils.set_field("reserved"), name="remove_resource_reserved"),
    url(r'^api/elrados/lock_resource/(?P<data_name>.*?)/?$',
        utils.lock_resource, name="lock_resource"),
    url(r'^/?$', "elrados.views.index"),
]



""""Set a field of resource module."""
import json
from socket import gethostbyaddr

from django.http.response import JsonResponse
from rotest.management.models.resource_data import ResourceData


def get_client_ip(request):
    """Get client's ip address."""
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")  # forward ip
    if x_forwarded_for:
        ip_address = x_forwarded_for.split(",")[0]

    else:
        ip_address = request.META.get("REMOTE_ADDR")

    return ip_address


def lock_resource(request, data_name):
    """"Lock a specific given resource."""
    lock_function = set_field("reserved")
    user = gethostbyaddr(get_client_ip(request))[0]
    data = json.loads(lock_function(request, data_name, user).content)
    data["user"] = user
    return JsonResponse(data)


def set_field(field_name):
    """"Set a field recursively.

    Args:
        field_name (str): the field to set.

    Returns:
        function. a function that allows to set a value of a given field.
    """

    # pylint: disable=unused-argument
    def set_resource_field(request, data_name, value=""):
        """"Entry point for removing any field from a resource.

        Args:
            request (HttpRequest): GET http request.
            data_name (str): the name of the resource.
            value (str): the value of the field to set to
                (default: "" - clear field).

        Returns:
            JsonResponse. the components that were changed.
        """
        # pylint: disable=no-member
        effected_resource_head = \
            ResourceData.objects.filter(name=data_name)[0]
        effected_resources = _field_set_recursively(effected_resource_head,
                                                    field_name,
                                                    value)

        return JsonResponse({"effected_resources": effected_resources})

    return set_resource_field


def _field_set_recursively(resource, field_name, value):
    """Set a field for a given resource and all of its children.

    Args:
        resource (ResourceData): the resource to change.
        field_name (str): the field to set.
        value (str): the value of the field to set to.

    Returns:
        list. all the effected resources.
    """
    effected_resources = []
    for sub_resource in resource.leaf.get_sub_resources():
        set_resources = _field_set_recursively(sub_resource,
                                               field_name,
                                               value)
        effected_resources.extend(set_resources)

    setattr(resource, field_name, value)
    resource.save()

    effected_resources.append(resource.name)
    return effected_resources



"""Setup file for handling packaging and distribution."""
import sys

from setuptools import setup, find_packages

__version__ = "0.6.2"

requirements = [
    'django>=1.8,<1.9',
    'twisted',
    'crochet',
    'autobahn',
    'rotest>=5.5.0',
]

setup(
    name='elrados',
    version=__version__,
    description="Frontend for django models and Rotest Testing Framework",
    license="MIT",
    author="Elran Shefer",
    author_email="elran777@gmail.com",
    url="https://github.com/IamShobe/elrados",
    keywords="rotest frontend",
    install_requires=requirements,
    python_requires="~=2.7.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"elrados": ["static/*",
                              "static/img/*",
                              "static/font/*",
                              "templates/*"]},
    zip_safe=False,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        ],
    entry_points={
        "rotest.cli_server_actions":
            ["elrados_server = elrados.backend.main:setup_server"],
    },
)
